#!/usr/bin/python

# auto-logout: enforces time limits for computer usage.
# Allowed login times and a maximal usage time for each user can be
# configured (see CONFIGURATION below).
# Uses systemd's login1 D-Bus interface to track the X11 session of each
# user; remote or text console logins are ignored and should be disabled.
# On a PC, users can have exactly one X11 session.
# Uses SessionManager D-Bus interface to logout users.
# Uses UPower D-Bus interface to track suspend/resume.   

import dbus, os, pwd, gobject, time, glib
from dbus.mainloop.glib import DBusGMainLoop
from datetime import datetime, timedelta
import pickle, signal

DEBUG = True

# global variables and constants
users = {}      # user limits and login times, indexed by UID
tracked_sessions = {}     # uids of tracked_sessions, indexed by session path
store_path = '/var/lib/autologout'   # store for user objects


# helper functions

def get_uname(uid):
# returns user name for UID
	try:
		user_pw = pwd.getpwuid(uid)
		return(user_pw[0])
	except:
		return 'guest_session'

def get_uid(uname):
# returns UID for user name
	user_pw = pwd.getpwnam(uname)
	return(user_pw[2])

def seconds(s):
# converts a string 'hh:mm[:ss]' to seconds since midnight
	if ':' in s:
		try:
			h, m, sec = s.split(':', 2)
			return(int(h)*3600 + int(m)*60 + int(sec))
		except:
			h, m = s.split(':', 1)
			return(int(h)*3600 + int(m)*60)
	else:
		return(0)

def time_str(seconds):
# converts seconds to string 'hh:mm:ss'
	h = seconds / 3600
	m = (seconds - h*3600) / 60
	s = seconds - h*3600 - m*60
	if h < 10:
		h = '0' + str(h)
	else:
		h = str(h)
	if m < 10:
		m = '0' + str(m)
	else:
		m = str(m)
	if s < 10:
		s = '0' + str(s)
	else:
		s = str(s)
	return h + ':' + m + ':' + s
	 
def now():
# returns the time in seconds since midnight
	dt = datetime.now()
	tm = dt.timetuple()
	return tm.tm_sec + 60*tm.tm_min + 3600*tm.tm_hour

def write_user(uid):
# writes recent state of user object for UID to disk
	f = open(store_path + '/' + get_uname(uid), 'w')
	pickle.dump(users[uid], f)
	f.close()

def read_user(uid):
# writes user object for UID to disk 
	try:
		f = open(store_path + '/' + get_uname(uid), 'r')
		users[uid] = pickle.load(f)
		f.close()
	except:
		pass

def get_session_address(uid):
# returns DBUS address of user UID's session or ''
	address = ''
	# search in /proc for the gnome-session process of user UID, 
	for path in os.listdir('/proc/'):
		fname = '/proc/' + path + '/exe'
		if os.path.exists(fname):
			stat = os.lstat(fname)
			if stat.st_uid == uid:
				if 'gnome-session' in os.readlink(fname):
				# find session bus address in environment
					f = open('/proc/' + path + '/environ', 'rb')
					environ = f.read()
					env = environ.split('\0')
					for e in env:
						if '=' in e:
							var, val = e.split('=', 1)
						if 'DBUS_SESSION_BUS_ADDRESS' == var:
							address = val
					f.close()
	return address


def logout(uid):
# starts Logout in gnome session for user with given uid
	time.sleep(5)                # wait for session to start up
	address = get_session_address(uid)
	# connect to users's session bus; call session manager's Logout function
	if address != '':
		os.seteuid(uid)      # root isn't allowed to write to uid's session bus
		try:
			session_bus = dbus.bus.BusConnection(address)
			if DEBUG: print '>>> logout()', address
			session_manager = session_bus.get_object('org.gnome.SessionManager',
			'/org/gnome/SessionManager')
			i = dbus.UInt32(0)
			session_manager.Logout(i)    # logout with cancel option
			os.seteuid(0)
		except:
			print('Error: could not get session manager for user', get_uname(uid))
			os.seteuid(0)
			return False
		time.sleep(60)                        # wait one minutes for user to stop his work
		try:                                  # user might be logged out meanwhile
			users[uid].deactivate()       # hard kill: no deactivate signal is sent
			i = dbus.UInt32(2)
			os.seteuid(uid)
			session_manager.Logout(i)     # logout immediately
			os.seteuid(0)
		except:
			os.seteuid(0)
			return False
	else:
		print('Error: could not get session dbus for user', get_uname(uid))
	return False   # needed to stop timer



class User:
# manages the limits and session status of one user with defined limits
	def __init__(self, name='', uid=0, start='0:00', end='24:00', limit=''):
		self.start = seconds(start)   # start of allowed use interval
		self.end = seconds(end)       # end of allowed use interval
		self.limit = seconds(limit)   # length of allowed use interval
		try:                          # don't die in case of non-existung user name oder UID
			if name != '':           	  # user name and UID
				self.uname = name         # if both are given, name overrides UID
				self.uid = get_uid(name)
			else:
				self.uid = uid
				self.uname = get_uname(uid)
		except:
			raise ValueError
		self.active_since = -1        # time session was activated, set by activate()
		self.cumulated = 0            # cumulates seconds over active periods
		self.last_reset = datetime.now() # last reset of cumulated useage time
		gobject.timeout_add_seconds(24*60*60, self.reset)


	def print_user(self):
		print self.uname + '(' + str(self.uid) + '): from', 
		print time_str(self.start), 'to', time_str(self.end), 
		print time_str(self.cumulated), 'from', time_str(self.limit), 'hours'
		

	def reset(self):
	# useage data are resetted each 24 hours
		self.cumulated = 0
		self.last_reset = datetime.now()
		write_user(self.uid)


	def activate(self):
	# user's X11 session is started, detected at program startup,
	# activated after a fast user switch or after a resume from susped/hibernate
		if DEBUG: print '>>> activate()', self.uname
		self.active_since = now()
		write_user(self.uid)
		delta = datetime.now() - self.last_reset
		if delta.days >= 1:    # at least 24 hours have passed since last reset
			self.reset()
		if self.active_since >= self.end:       # too late
			logout(self.uid)
		elif self.active_since <= self.start:     # too early
			logout(self.uid)
		elif self.limit > 0:                     # user has a limit
			if self.cumulated >= self.limit:   # limit exceeded
				logout(self.uid)
			else:                              # trigger terminate
				gobject.timeout_add_seconds(self.limit - self.cumulated,
				self.terminate)


	def deactivate(self):
	# user session has been deactivated or ended, or KeyboardInterrupt occured
		if DEBUG: print '>>> deactivate()', self.uname
		if self.active_since > -1:               # session is active
			if now() >= self.active_since:
				self.cumulated += now() - self.active_since
			else:                         # midnight has passed
				self.cumulated += (24*60*60 - self.active_since) + now()
			self.active_since = -1
		write_user(self.uid)


	def terminate(self, hard_kill=False):
	# triggered when time limit might be reached
		if self.active_since > -1:    # session is active
			if now() >= self.active_since:
				all_cumulated = self.cumulated + (now() - self.active_since)
			else:         # midnight has passed
				all_cumulated = self.cumulated + (24*60*60 - self.active_since) + now()
			if all_cumulated > self.limit - 10:   # close to limit:
				logout(self.uid, hard_kill)              # terminate now
			else:                                 # trigger terminateTerminate
				gobject.timeout_add_seconds(self.limit - all_cumulated,
				self.terminate)
		else:                         # session is inactive:
			pass                  # will be checked on next activation
		return False

# end class User



# callback function for PorpertiesChanged signal of sessions

def active_changed(interface, changed_props, invalidated_props):
	for session_path in tracked_sessions:
		session = bus.get_object('org.freedesktop.login1', session_path)
		active = session.Get('org.freedesktop.login1.Session', 'Active')
		uid = tracked_sessions[session_path]
		if DEBUG:
			print get_uname(uid),
			if active: print 'is active'
			else: print 'is inactive'
		if active: users[uid].activate()
		else: users[uid].deactivate()
	if DEBUG:
		for uid in users.keys():
			u = users[uid]
			print u.uname, u.cumulated, 'sec since', u.active_since, 'reset', u.last_reset
		print




# callback functions for login1 signals

def session_added(session_id, session_path):
# new session added; 
# session_id is something like 'c3', session_path something like '/org/freedesktop/login1/c3'
# non-X11 sessions (remote or text console logins) are ignored,
# guest sessions are terminated immediately
	if DEBUG: print 'session_added()', session_path, time_str(now())
	session = bus.get_object('org.freedesktop.login1', session_path)
	uid = session.Get('org.freedesktop.login1.Session', 'User')[0]
	if DEBUG: print 'User', get_uname(uid)
	if 'guest' in get_uname(uid):    # guest sessions are not allowed
		session.Terminate()
	if uid in users.keys():         # User object for session owner exists
		tracked_sessions[session_path] = uid
		session.connect_to_signal('PropertiesChanged', active_changed)
		users[uid].activate()
#	except:
#		if DEBUG:
#			print '>>> something strange happend with session', session_path


def session_removed(session_id, session_path):
# session removed
	if DEBUG: print 'session_removed()', session_path, time_str(now())
	if session_path in tracked_sessions:
		if DEBUG: print 'User', get_uname(tracked_sessions[session_path])
		users[tracked_sessions[session_path]].deactivate()
		del tracked_sessions[session_path]


def sleeping():
# system is going to suspend or hibernate
	for session_path in tracked_sessions:
		session = bus.get_object('org.freedesktop.login1', session_path)
		if session.Get('org.freedesktop.login1.Session', 'Active'):    # active session
			uid = tracked_sessions[session_path]
			users[uid].deactivate()

def resuming():
# system is resuming from suspend or hibernate
	for session_path in tracked_sessions:      # check tracked sessions
		session = bus.get_object('org.freedesktop.login1', session_path)
		if session.Get('org.freedesktop.login1.Session', 'Active'):    # active session
			uid = tracked_sessions[session_path]
			users[uid].activate()



# main

#################
# CONFIGURATION #
#################

# user limits, indexed by UID; adapt to your own needs. User objects are created.
# ATTENTION: after changes, remove status files in /var/lib/autologout!

# foo: two and a half hours between 14:00 and 20:00
#users[get_uid('foo')] = User(name='foo', start='14:00', end='20:00', limit='2:30')

# bar: no limits
#users[get_uid('bar')] = User(name='bar')

#users[get_uid('oliver')] = User(name='oliver', start='0:00', end='24:00', limit='12:00')
#users[get_uid('anna')] = User(name='anna', start='14:00', end='20:00', limit='0:01')
#users[get_uid('malin')] = User(name='malin', start='14:00', end='14:30')



# get dumped data from disk, if available
# overwrites limits; so remove status files, if configuration is changed
for uid in users.keys():
	read_user(uid)
	if DEBUG: users[uid].print_user()
if DEBUG: print

# root privilege needed for some needed functions of login1 service
if os.geteuid() != 0:
	print 'root privileges needed, sorry!'

else:
# create directory for storage of user objects
	if not os.path.exists(store_path):
		os.mkdir(store_path)

# set up a DBus event loop
	dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
# connect to the system bus
	bus = dbus.SystemBus() 
# create a proxy object for login1 Manager interface
	login1 = bus.get_object('org.freedesktop.login1', '/org/freedesktop/login1')
# scan existing sessionsS
	for session in login1.ListSessions():
		session_added(session[0], session[4])    # id and login1 path of the session

# callback functions for signals SessionNew and SessionRemoved
	login1.connect_to_signal('SessionNew', session_added)
	login1.connect_to_signal('SessionRemoved', session_removed)

# connect to UPower and catch signals Sleeping and Resuming
# only works with upower version >= 0.9.4 (Ubuntu 11.04 and above)
	upower = bus.get_object('org.freedesktop.UPower', '/org/freedesktop/UPower')
	upower.connect_to_signal('Sleeping', sleeping)
	upower.connect_to_signal('Resuming', resuming)

# enter event loop, catch KeyboardInterrupt
	try:
		loop = gobject.MainLoop()
		loop.run()
	except:   # Ctrl+C, kill
		for uid in users.keys():
			users[uid].deactivate()   # update user objects

